import torch
import torch.nn as nn
import torch.nn.functional as F
from hybra.utils import audfilters_fir, fir_tightener3000
from hybra.utils import plot_response as plot_response_
from hybra._fit_neurodual import fit

class AudletFIR(nn.Module):
    def __init__(self, filterbank_config={'filter_len':256,
                                          'num_channels':64,
                                          'fs':16000,
                                          'Ls':16000,
                                          'bwmul':1},
                                          is_encoder_learnable=False,
                                          use_decoder=False,
                                          is_decoder_learnable=False,
                                          decoder_fit_eps=1e-5):
        super().__init__()

        [filters, d, fc, fc_crit, L] = audfilters_fir(**filterbank_config)

        self.filters = filters
        self.stride = d
        self.filter_len = filterbank_config['filter_len'] 
        self.fs = filterbank_config['fs']
        self.fc = fc
        self.fc_crit = fc_crit

        kernels_real = torch.tensor(filters.real, dtype=torch.float32)
        kernels_imag = torch.tensor(filters.imag, dtype=torch.float32)

        if is_encoder_learnable:
            self.register_parameter('kernels_real', nn.Parameter(kernels_real, requires_grad=True))
            self.register_parameter('kernels_imag', nn.Parameter(kernels_imag, requires_grad=True))
        else:
            self.register_buffer('kernels_real', kernels_real)
            self.register_buffer('kernels_imag', kernels_imag)
        
        self.use_decoder = use_decoder
        if use_decoder:
            max_iter = 1000 # TODO: should we do something like that?
            decoder_kernels_real, decoder_kernels_imag, _, _ = fit(filterbank_config, decoder_fit_eps, max_iter)

            if is_decoder_learnable:
                self.register_parameter('decoder_kernels_real', nn.Parameter(decoder_kernels_real, requires_grad=True))
                self.register_parameter('decoder_kernels_imag', nn.Parameter(decoder_kernels_imag, requires_grad=True))
            else:        	
                self.register_buffer('decoder_kernels_real', decoder_kernels_real)
                self.register_buffer('decoder_kernels_imag', decoder_kernels_imag)

    def forward(self, x):
        x = F.pad(x.unsqueeze(1), (self.filter_len//2, self.filter_len//2), mode='circular')

        out_real = F.conv1d(x, self.kernels_real.to(x.device).unsqueeze(1), stride=self.stride)
        out_imag = F.conv1d(x, self.kernels_imag.to(x.device).unsqueeze(1), stride=self.stride)

        return out_real + 1j * out_imag

    def decoder(self, x_real:torch.Tensor, x_imag:torch.Tensor) -> torch.Tensor:
        """Filterbank synthesis.

        Parameters:
        -----------
        x (torch.Tensor) - input tensor of shape (batch_size, n_filters, signal_length//hop_length)

        Returns:
        --------
        x (torch.Tensor) - output tensor of shape (batch_size, signal_length)
        """
        x = (
            F.conv_transpose1d(
                x_real,
                self.decoder_kernels_real.to(x_real.device).unsqueeze(1),
                stride=self.stride,
                padding=self.filter_len//2,
                output_padding=self.stride - 2
            ) + F.conv_transpose1d(
                x_imag,
                self.decoder_kernels_imag.to(x_imag.device).unsqueeze(1),
                stride=self.stride,
                padding=self.filter_len//2,
                output_padding=self.stride - 2
            )
        )

        return x.squeeze(1)

    def plot_response(self):
        plot_response_(g=(self.kernels_real + 1j*self.kernels_imag).detach().numpy(), fs=self.fs, scale=True, fc_crit=self.fc_crit)

    def plot_decoder_response(self):
        if self.use_decoder:
            plot_response_(g=(self.decoder_kernels_real+1j*self.decoder_kernels_imag).detach().numpy(), fs=self.fs, decoder=True)
        else:
            raise NotImplementedError("No decoder configured")



import torch
import torch.nn as nn
import torch.nn.functional as F
from hybra.utils import calculate_condition_number, fir_tightener3000, audfilters_fir, plot_response

class HybrA(nn.Module):
    def __init__(self, filterbank_config={'filter_len':128,
                                          'num_channels':64,
                                          'fs':16000,
                                          'Ls':16000,
                                          'bwmul':1}, kernel_len=24, start_tight:bool=True):
        
        super().__init__()

        [filters, d, fc, fc_crit, L] = audfilters_fir(**filterbank_config)

        self.filters = filters
        self.stride = d
        self.filter_len = filterbank_config['filter_len']
        self.kernel_len = kernel_len
        self.fs = filterbank_config['fs']
        self.fc = fc
        self.fc_crit = fc_crit
        self.num_channels = filters.shape[0]

        fir_kernels_real = torch.tensor(filters.real, dtype=torch.float32)
        fir_kernels_imag = torch.tensor(filters.imag, dtype=torch.float32)

        self.register_buffer('fir_kernels_real', fir_kernels_real)
        self.register_buffer('fir_kernels_imag', fir_kernels_imag)
        self.output_real_forward = None
        self.output_imag_forward = None

        # Initialize trainable filters
        k = torch.tensor(self.num_channels / (self.kernel_len * self.num_channels))
        encoder_weight = (-torch.sqrt(k) - torch.sqrt(k)) * torch.rand([self.num_channels, 1, self.kernel_len]) + torch.sqrt(k)

        if start_tight:
            encoder_weight = torch.tensor(fir_tightener3000(
                encoder_weight.squeeze(1), self.kernel_len, D=d, eps=1.01
            ),  dtype=torch.float32).unsqueeze(1)
            encoder_weight = encoder_weight / torch.norm(encoder_weight, dim=-1, keepdim=True)
        
        self.encoder_weight_real = nn.Parameter(encoder_weight, requires_grad=True)
        self.encoder_weight_imag = nn.Parameter(encoder_weight, requires_grad=True)

        # compute the initial hybrid filters
        self.hybra_filters_real = F.conv1d(
            self.fir_kernels_real.squeeze(1),
            self.encoder_weight_real,
            groups=self.num_channels,
            padding="same",
        ).unsqueeze(1)
        self.hybra_filters_imag = F.conv1d(
            self.fir_kernels_imag.squeeze(1),
            self.encoder_weight_imag,
            groups=self.num_channels,
            padding="same",
        ).unsqueeze(1)

    def forward(self, x:torch.Tensor) -> torch.Tensor:
        """Forward pass of the HybridFilterbank.
        
        Parameters:
        -----------
        x (torch.Tensor) - input tensor of shape (batch_size, 1, signal_length)
        
        Returns:
        --------
        x (torch.Tensor) - output tensor of shape (batch_size, num_channels, signal_length//hop_length)
        """

        kernel_real = F.conv1d(
            self.fir_kernels_real.to(x.device).squeeze(1),
            self.encoder_weight_real,
            groups=self.num_channels,
            padding="same",
        ).unsqueeze(1)
        self.hybra_filters_real = kernel_real.clone().detach()

        kernel_imag = F.conv1d(
            self.fir_kernels_imag.to(x.device).squeeze(1),
            self.encoder_weight_imag,
            groups=self.num_channels,
            padding="same",
        ).unsqueeze(1)
        self.hybra_filters_imag = kernel_imag.clone().detach()
        
        output_real = F.conv1d(
            F.pad(x.unsqueeze(1), (self.filter_len//2, self.filter_len//2), mode='circular'),
            kernel_real,
            stride=self.stride,
        )
        
        output_imag = F.conv1d(
            F.pad(x.unsqueeze(1), (self.filter_len//2,self.filter_len//2), mode='circular'),
            kernel_imag,
            stride=self.stride,
        )

        return output_real + 1j* output_imag

    def encoder(self, x:torch.Tensor):
        """For learning use forward method

        """
        out = F.conv1d(
                    F.pad(x.unsqueeze(1),(self.filter_len//2, self.filter_len//2), mode='circular'),
                    self.hybra_filters_real.to(x.device),
                    stride=self.stride,
                ) + 1j * F.conv1d(
                    F.pad(x.unsqueeze(1),(self.filter_len//2, self.filter_len//2), mode='circular'),
                    self.hybra_filters_imag.to(x.device),
                    stride=self.stride,
                )
                
        return out
    
    def decoder(self, x_real:torch.Tensor, x_imag:torch.Tensor) -> torch.Tensor:
        """Forward pass of the dual HybridFilterbank.

        Parameters:
        -----------
        x (torch.Tensor) - input tensor of shape (batch_size, num_channels, signal_length//hop_length)

        Returns:
        --------
        x (torch.Tensor) - output tensor of shape (batch_size, signal_length)
        """
        x = (
            F.conv_transpose1d(
                x_real,
                self.hybra_filters_real,
                stride=self.stride,
                padding=self.filter_len//2,
            )
            + F.conv_transpose1d(
                x_imag,
                self.hybra_filters_imag,
                stride=self.stride,
                padding=self.filter_len//2,
            )
        )

        return x.squeeze(1)

    # @property
    # def condition_number(self):
    #     # coefficients = self.hybra_filters_real.detach().clone().squeeze(1) + 1j* self.hybra_filters_imag.detach().clone().squeeze(1)
    #     filters = (self.hybra_filters_real + 1j*self.hybra_filters_imag).squeeze()
    #     return calculate_condition_number(filters, self.stride)
    def plot_response(self):
        plot_response((self.hybra_filters_real + 1j*self.hybra_filters_imag).squeeze().detach().numpy(), self.fs)
    def plot_decoder_response(self):
        plot_response((self.hybra_filters_real + 1j*self.hybra_filters_imag).squeeze().detach().numpy(), self.fs, decoder=True)



import numpy as np
import torch
import matplotlib.pyplot as plt

####################################################################################################
##################### Cool routines to study decimated filterbanks #################################
####################################################################################################

def condition_number(w_hat:torch.Tensor, D:int) -> torch.Tensor:
    """
    Computes the condition number of a filterbank w_hat using the polyphase representation.
    Input:  w: Frequency responses of the filterbank as 2-D Tensor torch.tensor[length, num_channels]
            D: Decimation (or downsampling) factor, must divide filter length!
    Output: Condition number.
    """
    if D == 1:
        lp = torch.sum(w_hat.abs() ** 2, dim=1)
        A = torch.min(lp)
        B = torch.max(lp)
        return B/A
    else:    
        N = w_hat.shape[0]
        J = w_hat.shape[1]
        assert N % int(D) == 0, "Oh no! Decimation factor must divide signal length!"

        A = torch.tensor([torch.inf]).to(w_hat.device)
        B = torch.tensor([0]).to(w_hat.device)
        Ha = torch.zeros((D,J)).to(w_hat.device)
        Hb = torch.zeros((D,J)).to(w_hat.device)

        for j in range(N//D):
            idx_a = (j - torch.arange(D) * (N//D)) % N
            idx_b = (torch.arange(D) * (N//D) - j) % N
            Ha = w_hat[idx_a, :]
            Hb = torch.conj(w_hat[idx_b, :])
            lam = torch.linalg.eigvalsh(Ha @ Ha.H + Hb @ Hb.H).real
            A = torch.min(A, torch.min(lam))
            B = torch.max(B, torch.max(lam))
        return B/A

def calculate_condition_number(w:torch.Tensor, D:int):
    """
    in: frequency responses of fb (n_filters, system length), decimation factor
    out: condition number
    """
    w_hat = torch.fft.fft(w, dim=-1).T
    return condition_number(w_hat, D)

def can_tight(w:torch.Tensor, D:int) -> torch.Tensor:
    """
    Computes the canonical tight filterbank of w (time domain) using the polyphase representation.
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, length]
            D: Decimation (or downsampling) factor, must divide filter length!
    Output: Canonical tight filterbank of W (torch.tensor[num_channels, length])
    """
    w_hat = torch.fft.fft(w.T, dim=0)
    if D == 1:
        lp = torch.sum(w_hat.abs() ** 2, dim=1).reshape(-1,1)
        w_hat_tight = w_hat * (lp ** (-0.5))
        return torch.fft.ifft(w_hat_tight.T, dim=1)
    else:
        N = w_hat.shape[0]
        J = w_hat.shape[1]
        assert N % int(D) == 0, "Oh no! Decimation factor must divide signal length!"

        w_hat_tight = torch.zeros(J, N, dtype=torch.complex64)
        for j in range(N//D):
            idx = (j - torch.arange(D) * (N//D)) % N
            H = w_hat[idx, :]
            U, _, V = torch.linalg.svd(H, full_matrices=False)
            H = U @ V
            w_hat_tight[:,idx] = H.T.to(torch.complex64)
        return torch.fft.ifft(torch.fft.ifft(w_hat_tight.T, dim=1) * D ** 0.5, dim=0).T

def frame_bounds(w, D):
    """
    in: frequency responses of fb (system length, n_filters), decimation factor
    out: frame bounds
    """
    N = w.shape[0]
    M = w.shape[1]
    assert N % D == 0

    A = torch.inf
    B = 0
    Ha = torch.zeros((D,M))
    Hb = torch.zeros((D,M))

    for j in range(N//D):
        idx_a = np.mod(j - np.arange(D) * (N//D), N).astype(int)
        idx_b = np.mod(np.arange(D) * (N//D) - j, N).astype(int)
        Ha = w[idx_a, :]
        Hb = torch.conj(w[idx_b, :])
        lam = torch.linalg.eigvalsh(Ha @ Ha.H + Hb @ Hb.H).real
        A = np.min([A, torch.min(lam).item()]).item()
        B = np.max([B, torch.max(lam).item()]).item()
    return A/D, B/D

def kappa_alias(w:torch.Tensor, D:int) -> torch.Tensor:
    """
    Computes the frequency correlation functions.
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, sig_length]
            D: Decimation (or downsampling) factor, must divide filter length!
            aliasing: If False, only the condition umber is returned
    Output: Condition number and norm of the aliasing term
    """
    w_hat = torch.fft.fft(w, dim=-1).T
    N = w_hat.shape[0]
    G = torch.zeros(N, D)
    G[:,0] = torch.sum(torch.abs(w_hat)**2, dim=1)
    for j in range(1,D):
        G[:,j] = torch.sum(w_hat * torch.conj(w_hat.roll(j * N//D, 0)), dim=1)
    return G

def alias_conditioner(w:torch.Tensor, D:int, aliasing:bool=True) -> torch.Tensor:
    G = kappa_alias(w, D)
    return torch.max(G[:,0]).div(torch.min(G[:,0])) + torch.norm(G[:,1::], dim=1)**2 - 1

def fir_tightener3000(w, supp, D, eps=1.01, Ls=None):
    """
    Iterative tightening procedure with fixed support for a given filterbank w
    Input:  w: Impulse responses of the filterbank as 2-D Tensor torch.tensor[num_channels, length].
            supp: Desired support of the resulting filterbank
            D: Decimation (or downsampling) factor, must divide filter length!
            eps: Desired condition number
            Ls: control syste length
    Output: Filterbank with condition number *kappa* and support length *supp*. If length=supp then the resulting filterbank is the canonical tight filterbank of w.
    """
    if Ls is not None:
        w =  torch.cat([w, torch.zeros(w.shape[0], Ls-w.shape[1])], dim=1)
    w_tight = w.clone()
    kappa = calculate_condition_number(w, D).item()
    while kappa > eps:
        w_tight = can_tight(w_tight, D)
        w_tight[:, supp:] = 0
        kappa = calculate_condition_number(w_tight, D).item()
    if Ls is None:
        return w_tight
    else:
        return w_tight[:,:supp]
    

####################################################################################################
################### Routines for constructing auditory filterbanks #################################
####################################################################################################

def freqtoaud(freq, scale="erb"):
    """
    Converts frequencies (Hz) to auditory scale units.

    Parameters:
    freq (float or ndarray): Frequency value(s) in Hz.
    scale (str): Auditory scale. Supported values are:
                 - 'erb' (default)
                 - 'mel'
                 - 'bark'
                 - 'log10'

    Returns:
    float or ndarray: Corresponding auditory scale units.
    """

    scale = scale.lower()
    
    if scale == "mel":
        # MEL scale
        return 1000 / np.log(17 / 7) * np.sign(freq) * np.log(1 + np.abs(freq) / 700)

    elif scale == "erb":
        # Glasberg and Moore's ERB scale
        return 9.2645 * np.sign(freq) * np.log(1 + np.abs(freq) * 0.00437)

    elif scale == "bark":
        # Bark scale from Traunmuller (1990)
        return np.sign(freq) * ((26.81 / (1 + 1960 / np.abs(freq))) - 0.53)

    elif scale in ["log10", "semitone"]:
        # Logarithmic scale
        return np.log10(freq)

    else:
        raise ValueError(f"Unsupported scale: '{scale}'. Available options are: 'mel', 'erb', 'bark', 'log10'.")

def audtofreq(aud, scale="erb"):
    """
    Converts auditory units to frequency (Hz).
    Parameters:
    aud (float or numpy array): Auditory scale value(s) to convert.
    scale (str): Auditory scale. Supported values are:
                 - 'erb' (default)
                 - 'mel'
                 - 'bark'
                 - 'log10'

    Returns:
    float or numpy array: Frequency value(s) in Hz.
    """
    if scale == "mel":
        return 700 * np.sign(aud) * (np.exp(np.abs(aud) * np.log(17 / 7) / 1000) - 1)
    elif scale == "erb":
        return (1 / 0.00437) * (np.exp(aud / 9.2645) - 1)
    elif scale == "bark":
        return np.sign(aud) * 1960 / (26.81 / (np.abs(aud) + 0.53) - 1)
    elif scale in ["log10", "semitone"]:
        return 10 ** aud
    else:
        raise ValueError(f"Unsupported scale: '{scale}'. Available options are: 'mel', 'erb', 'bark', 'log10'.")


def audspace(fmin, fmax, num_channels, scale="erb"):
    """
    Computes a vector of values equidistantly spaced on the selected auditory scale.

    Parameters:
    fmin (float): Minimum frequency in Hz.
    fmax (float): Maximum frequency in Hz.
    num_channels (int): Number of points in the output vector.
    audscale (str): Auditory scale (default is 'erb').

    Returns:
    tuple:
        y (ndarray): Array of frequencies equidistantly scaled on the auditory scale.
        bw (float): Bandwidth between each sample on the auditory scale.
    """
    if not (isinstance(fmin, (int, float)) and np.isscalar(fmin)):
        raise ValueError("fmin must be a scalar.")
    
    if not (isinstance(fmax, (int, float)) and np.isscalar(fmax)):
        raise ValueError("fmax must be a scalar.")
    
    if not (isinstance(num_channels, int) and num_channels > 0):
        raise ValueError("n must be a positive integer scalar.")
    
    if fmin > fmax:
        raise ValueError("fmin must be less than or equal to fmax.")

    # Convert [fmin, fmax] to auditory scale
    audlimits = freqtoaud(np.array([fmin, fmax]), scale)

    # Generate frequencies spaced evenly on the auditory scale
    aud_space = np.linspace(audlimits[0], audlimits[1], num_channels)
    y = audtofreq(aud_space, scale)

    # Ensure exact endpoints
    y[0] = fmin
    y[-1] = fmax

    return y

def freqtoaud_mod(freq, fc_crit):
    """
    Modified auditory scale function with linear region below fc_crit.
    
    Parameters:
    freq (ndarray): Frequency values in Hz.
    fc_crit (float): Critical frequency in Hz.

    Returns:
    ndarray: Values on the modified auditory scale.
    """
    aud_crit = freqtoaud(fc_crit)
    slope = (freqtoaud(fc_crit * 1.01) - aud_crit) / (fc_crit * 0.01)

    aud = np.zeros_like(freq, dtype=np.float32)
    linear_part = freq < fc_crit
    auditory_part = freq >= fc_crit

    aud[linear_part] = slope * (freq[linear_part] - fc_crit) + aud_crit
    aud[auditory_part] = freqtoaud(freq[auditory_part])

    return aud

def audtofreq_mod(aud, fc_crit):
    """
    Inverse of freqtoaud_mod to map auditory scale back to frequency.
    
    Parameters:
    aud (ndarray): Auditory scale values.
    fc_crit (float): Critical frequency in Hz.

    Returns:
    ndarray: Frequency values in Hz
    """
    aud_crit = freqtoaud(fc_crit)
    slope = (freqtoaud(fc_crit * 1.01) - aud_crit) / (fc_crit * 0.01)

    freq = np.zeros_like(aud, dtype=np.float32)
    linear_part = aud < aud_crit
    auditory_part = aud >= aud_crit

    freq[linear_part] = (aud[linear_part] - aud_crit) / slope + fc_crit
    freq[auditory_part] = audtofreq(aud[auditory_part])

    return freq

def audspace_mod(fc_crit, fs, num_channels):
    """Generate M frequency samples that are equidistant in the modified auditory scale.
    
    Parameters:
    fc_crit (float): Critical frequency in Hz.
    fs (int): Sampling rate in Hz.
    M (int): Number of filters/channels.

    Returns:
    ndarray: Frequency values in Hz and in the auditory scale.
    """

    # Convert [0, fs//2] to modified auditory scale
    aud_min = freqtoaud_mod(np.array([0]), fc_crit)[0]
    aud_max = freqtoaud_mod(np.array([fs//2]), fc_crit)[0]

    # Generate frequencies spaced evenly on the modified auditory scale
    fc_aud = np.linspace(aud_min, aud_max, num_channels)

    # Convert back to frequency scale
    fc = audtofreq_mod(fc_aud, fc_crit)

    # Ensure exact endpoints
    fc[0] = 0
    fc[-1] = fs//2

    return fc, fc_aud

def fctobw(fc, scale="erb"):
    """
    Computes the critical bandwidth of a filter at a given center frequency.

    Parameters:
    fc (float or ndarray): Center frequency in Hz. Must be non-negative.
    audscale (str): Auditory scale. Supported values are:
                    - 'erb': Equivalent Rectangular Bandwidth (default)
                    - 'bark': Bark scale
                    - 'mel': Mel scale
                    - 'log10': Logarithmic scale

    Returns:
    ndarray or float: Critical bandwidth at each center frequency.
    """
    if isinstance(fc, (list, tuple)):
        fc = np.array(fc)
    if not (isinstance(fc, (float, int, np.ndarray)) and np.all(fc >= 0)):
        raise ValueError("fc must be a non-negative scalar or array.")

    # Compute bandwidth based on the auditory scale
    if scale == "erb":
        bw = 24.7 + fc / 9.265
    elif scale == "bark":
        bw = 25 + 75 * (1 + 1.4e-6 * fc**2)**0.69
    elif scale == "mel":
        bw = np.log(17 / 7) * (700 + fc) / 1000
    elif scale in ["log10"]:
        bw = fc
    else:
        raise ValueError(f"Unsupported auditory scale: {scale}")

    return bw

def bwtofc(bw, scale="erb"):
    """
    Computes the center frequency corresponding to a given critical bandwidth.

    Parameters:
    bw (float or ndarray): Critical bandwidth. Must be non-negative.
    scale (str): Auditory scale. Supported values are:
                 - 'erb': Equivalent Rectangular Bandwidth
                 - 'bark': Bark scale
                 - 'mel': Mel scale
                 - 'log10': Logarithmic scale

    Returns:
    ndarray or float: Center frequency corresponding to the given bandwidth.
    """
    if isinstance(bw, (list, tuple)):
        bw = np.array(bw)
    if not (isinstance(bw, (float, int, np.ndarray)) and np.all(bw >= 0)):
        raise ValueError("bw must be a non-negative scalar or array.")

    # Compute center frequency based on the auditory scale
    if scale == "erb":
        fc = (bw - 24.7) * 9.265
    elif scale == "bark":
        fc = np.sqrt(((bw - 25) / 75)**(1 / 0.69) / 1.4e-6)
    elif scale == "mel":
        fc = 1000 * (bw / np.log(17 / 7)) - 700
    elif scale in ["log10"]:
        fc = bw
    else:
        raise ValueError(f"Unsupported auditory scale: {scale}")

    return fc

def firwin(window_length, padding_length=None):
    """
    FIR window generation in Python.
    
    Parameters:
        window_length (int): Length of the window.
        padding_length (int): Length to which it should be padded.
        name (str): Name of the window.
        
    Returns:
        g (ndarray): FIR window.
    """

    if window_length % 2 == 0:
        x = np.concatenate([np.linspace(0, 0.5 - 1/window_length, window_length//2), np.linspace(-0.5, -1/window_length, window_length//2)])
    else:
        x = np.concatenate([np.linspace(0, 0.5 - 0.5/window_length, window_length//2), np.linspace(-0.5 + 0.5/window_length, -0.5/window_length, window_length//2)])
    x += window_length//2 / window_length

    # Hann window
    g = 0.5 + 0.5 * np.cos(2 * np.pi * x)
    
    # L1 Normalization
    g /= np.sum(np.abs(g))
    #g /= np.max(np.abs(g))

    if padding_length is None:
        if window_length % 2 == 0:
            return g
        else:
            return np.concatenate([g, np.zeros(1)])
    elif padding_length == window_length:
        return g
    elif padding_length > window_length:
        g_padded = np.concatenate([g, np.zeros(padding_length - len(g))])
        g_centered = np.roll(g_padded, (padding_length - len(g))//2)
        return g_centered
    else:
        raise ValueError("padding_length must be larger than window_length.")


def modulate(g, fc, fs):
    """Modulate a filters.
    
    Args:
        g (list of torch.Tensor): Filters.
        fc (list): Center frequencies.
        fs (int): Sampling rate.
    
    Returns:
        g_mod (list of torch.Tensor): Modulated filters.
    """
    Lg = len(g)
    g_mod = g * np.exp(2 * np.pi * 1j * fc * np.arange(Lg) / fs)
    return g_mod


####################################################################################################
########################################## the #####################################################
################################## filterbank generator ############################################
####################################################################################################


def audfilters_fir(filter_len, num_channels, fs, Ls, bwmul=1, scale='erb'):
    """
    Generate FIR filter kernel with length *filter_len* equidistantly spaced on auditory frequency scales.
    
    Parameters:
        filter_len (int): Length of the FIR filter.
        num_channels (int): Number of channels.
        fs (int): Sampling rate.
        Ls (int): Signal length.
        bwmul (float): Bandwidth multiplier.
        scale (str): Auditory scale.
    
    Returns:
        filters (list of torch.Tensor): Generated filters.
        a (list): Downsampling rates.
        fc (list): Center frequencies.
        L (int): Admissible signal length.
    """

    ####################################################################################################
    # Bandwidth conversion
    ####################################################################################################

    probeLs = 10000
    probeLg = 1000
    g_probe = firwin(probeLg, probeLs)
    
    # peak normalize
    gf_probe = np.fft.fft(g_probe) / np.max(np.abs(np.fft.fft(g_probe)))

    # compute ERB-type bandwidth of the prototype
    bw_conversion = np.linalg.norm(gf_probe)**2 * probeLg / probeLs / 4
    weird_factor = fs * 10.64
    
    ####################################################################################################
    # Center frequencies
    ####################################################################################################

    # get the bandwidth for the maximum admissible filter length and the associated center frequency
    fsupp_crit = bw_conversion / filter_len * weird_factor
    fc_crit = bwtofc(fsupp_crit / bwmul * bw_conversion)
    fc_crit_aud = audtofreq(fc_crit)

    [fc, fc_aud] = audspace_mod(fc_crit, fs, num_channels)
    num_lin = np.where(fc < fc_crit)[0].shape[0]

    ####################################################################################################
    # Frequency and time supports
    ####################################################################################################

    # frequency support for the auditory part
    fsupp = fctobw(fc[num_lin:]) / bw_conversion * bwmul

    # time support for the auditory part
    tsupp_lin = (np.ones(num_lin) * filter_len).astype(int)
    tsupp_aud = (np.round(bw_conversion / fsupp * weird_factor)).astype(int)
    tsupp = np.concatenate([tsupp_lin, tsupp_aud])

    # Maximal decimation factor (stride) to get a nice frame and accoring signal length
    d = np.floor(np.min(fs / fsupp)).astype(int)
    L = int(np.ceil(Ls / d) * d)

    ####################################################################################################
    # Generate filters
    ####################################################################################################

    g = np.zeros((num_channels, filter_len), dtype=np.complex128)

    g[0,:] = np.sqrt(d) * firwin(filter_len) #/ np.sqrt(2)
    g[-1,:] = np.sqrt(d) * modulate(firwin(tsupp[-1], filter_len), fs//2, fs) #/ np.sqrt(2)

    for m in range(1, num_channels - 1):
        g[m,:] = np.sqrt(d) * modulate(firwin(tsupp[m], filter_len), fc[m], fs)

    return g, d, fc, fc_crit, L

####################################################################################################
####################################################################################################
####################################################################################################
####################################################################################################

def response(g, fs):
    """Frequency response of the filters.
    
    Args:
        g (numpy.Array): Filters.
        fs (int): Sampling rate for plotting Hz.
        a (int): Downsampling rate.
    """
    Lg = g.shape[-1]
    num_channels = g.shape[0]
    g_long = np.concatenate([g, np.zeros((num_channels, int(fs) - Lg))], axis=1)
    G = np.abs(np.fft.fft(g_long, axis=1)[:,:fs//2])**2

    return G

def plot_response(g, fs, scale=False, fc_crit=None, decoder=False):
    """Frequency response of the filters.
    
    Args:
        g (numpy.Array): Filters.
        a (int): Downsampling rate.
        fs (int): Sampling rate for plotting Hz.
        fc_orig (numpy.Array): Original center frequencies.
        fc_low (numpy.Array): Center frequencies of the low-pass filters.
        fc_high (numpy.Array): Center frequencies of the high-pass filters.
        ind_crit (int): Index of the critical filter.
    """
    g_hat = response(g, fs)
    psd = np.sum(g_hat, axis=0)

    if scale:
        fig, ax = plt.subplots(3, 1, figsize=(12, 8), sharex=True)

        scale_id = 0
        fr_id = 1
        psd_id = 2

        num_channels = g.shape[0]
        filter_length = g.shape[1]
        freq_samples, aud_samples = audspace_mod(fc_crit, fs, num_channels)
        freqs = np.linspace(0, fs//2, fs//2)

        auds = freqtoaud_mod(freqs, fc_crit)

        ax[scale_id].scatter(freq_samples, freqtoaud_mod(freq_samples, fc_crit), color="black", label="Center frequencies", linewidths = 0.05)
        ax[scale_id].plot(freqs, auds, color='black')

        if fc_crit is not None:
            ax[scale_id].axvline(fc_crit, color='salmon', linestyle='--', label="Transition frequency", alpha=0.5)
            ax[scale_id].fill_betweenx(y=[auds[0]-1, auds[-1]*1.1], x1=0, x2=fc_crit, color='gray', alpha=0.25)
            ax[scale_id].fill_betweenx(y=[auds[0]-1, auds[-1]*1.1], x1=fc_crit, x2=fs//2, color='gray', alpha=0.1)

            ax[fr_id].fill_betweenx(y=[0, np.max(g_hat)*1.1], x1=0, x2=fc_crit, color='gray', alpha=0.25)
            ax[fr_id].fill_betweenx(y=[0, np.max(g_hat)*1.1], x1=fc_crit, x2=fs//2, color='gray', alpha=0.1)
            ax[psd_id].fill_betweenx(y=[0, np.max(psd)*1.1], x1=0, x2=fc_crit, color='gray', alpha=0.25)
            ax[psd_id].fill_betweenx(y=[0, np.max(psd)*1.1], x1=fc_crit, x2=fs//2, color='gray', alpha=0.1)

        ax[scale_id].set_xlim([0, fs//2])
        ax[scale_id].set_ylim([auds[0]-1, auds[-1]*1.1])
        #ax[scale_id].set_xlabel("Frequency (Hz)")
        text_x = fc_crit / 2
        text_y = auds[-1] - 2
        ax[scale_id].text(text_x, text_y, 'linear', color='black', ha='center', va='center', fontsize=12, alpha=0.75)
        ax[scale_id].text(text_x + fc_crit - 1, text_y, 'ERB', color='black', ha='center', va='center', fontsize=12, alpha=0.75)
        ax[scale_id].set_title(f"Modified Auditory Scale for {num_channels} filters of length {filter_length}")
        ax[scale_id].set_ylabel("Auditory Units")
        ax[scale_id].legend(loc='lower right')

    else:
        fig, ax = plt.subplots(2, 1, figsize=(10, 6), sharex=True)

        fr_id = 0
        psd_id = 1
    
    f_range = np.linspace(0, fs//2, fs//2)
    ax[fr_id].set_xlim([0, fs//2])
    ax[fr_id].set_ylim([0, np.max(g_hat)*1.1])
    ax[fr_id].plot(f_range, g_hat.T)
    if decoder:
        ax[fr_id].set_title('Frequency responses of the synthesis filters')
    if not decoder:
        ax[fr_id].set_title('Frequency responses of the analysis filters')
    #ax[fr_id].set_xlabel('Frequency [Hz]')
    ax[fr_id].set_ylabel('Magnitude')

    ax[psd_id].plot(f_range, psd)
    ax[psd_id].set_xlim([0, fs//2])
    ax[psd_id].set_ylim([0, np.max(psd)*1.1])
    ax[psd_id].set_title('Power spectral density')
    ax[psd_id].set_xlabel('Frequency [Hz]')
    ax[psd_id].set_ylabel('Magnitude')

    plt.show()


import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from hybra.utils import audfilters_fir
import warnings

class MSETight(torch.nn.Module):
    def __init__(self, beta: float = 0.0, fs: int = 16000):
        super().__init__()
        self.beta = beta
        self.loss = torch.nn.MSELoss()
        self.fs = fs

    def forward(self, preds, target, w=None):
        loss = self.loss(preds, target)
        Lg = w.shape[-1]
        num_channels = w.shape[0]
        w_long = torch.concatenate([w, torch.zeros((num_channels, self.fs - Lg))], axis=1)
        w_hat = torch.sum(torch.abs(torch.fft.fft(w_long, dim=1)[:, :self.fs//2])**2, dim=0)
        kappa = w_hat.max() / w_hat.min()
        
        return loss, loss + self.beta * (kappa - 1), kappa.item()

def noise_uniform(dur=1, fs=16000):
    N = int(dur * fs)
    X = torch.rand(N // 2 + 1) * 2 - 1
    
    X_full = torch.zeros(N, dtype=torch.cfloat)
    X_full[0:N//2+1] = X
    X_full[N//2+1:] = torch.conj(X[1:N//2].flip(0))
    
    x = torch.fft.ifft(X_full).real
    x = x / torch.max(torch.abs(x))
    
    return x.unsqueeze(0)

class NeuroDual(nn.Module):
	def __init__(self, filterbank_config):
		super().__init__()
		
		[filters, d, fc, fc_crit, _] = audfilters_fir(**filterbank_config)
		self.filters = filters
		self.stride = d
		self.filter_len = filterbank_config['filter_len'] 
		self.fs = filterbank_config['fs']
		self.fc = fc
		self.fc_crit = fc_crit
		
		self.register_buffer('kernels_real', torch.tensor(filters.real, dtype=torch.float32))
		self.register_buffer('kernels_imag', torch.tensor(filters.imag, dtype=torch.float32))

		self.register_parameter('kernels_decoder_real', nn.Parameter(torch.tensor(filters.real, dtype=torch.float32), requires_grad=True))
		self.register_parameter('kernels_decoder_imag', nn.Parameter(torch.tensor(filters.imag, dtype=torch.float32), requires_grad=True))

	def forward(self, x):
		L_out = x.shape[-1]
		
		x = F.pad(x.unsqueeze(1), (self.filter_len//2, self.filter_len//2), mode='circular')
		x_real = F.conv1d(x, self.kernels_real.to(x.device).unsqueeze(1), stride=self.stride)
		x_imag = F.conv1d(x, self.kernels_imag.to(x.device).unsqueeze(1), stride=self.stride)
		
		L_in = x_real.shape[-1]
		kernel_size = self.kernels_decoder_real.shape[-1]
		padding = self.filter_len // 2

		# L_out = (L_in -1) * stride - 2 * padding + dialation * (kernel_size - 1) + output_padding + 1 ; dialation = 1
		output_padding=L_out - (L_in -1)*self.stride + 2*padding - kernel_size

		x = F.conv_transpose1d(
			x_real,
			self.kernels_decoder_real.unsqueeze(1),
			stride=self.stride,
			padding=padding,
			output_padding=output_padding
			) + F.conv_transpose1d(
				x_imag,
				self.kernels_decoder_imag.unsqueeze(1),
				stride=self.stride,
				padding=padding,
				output_padding=output_padding
			)
		
		return x

def fit(filterbank_config, eps_loss, max_iter):
	model = NeuroDual(filterbank_config=filterbank_config)
	optimizer = optim.Adam(model.parameters(), lr=5e-4)
	criterion = MSETight(beta=1e-8)

	losses = []
	kappas = []	

	loss_item = float('inf')
	i = 0
	print("Computing the synthesis filterbank. This might take a while :)")
	while loss_item >= eps_loss:
		optimizer.zero_grad()
		x = noise_uniform(filterbank_config['Ls']/filterbank_config['fs'],filterbank_config['fs'])
		output = model(x)
		
		w_real = model.kernels_decoder_real.squeeze()
		w_imag = model.kernels_decoder_imag.squeeze()
		
		loss, loss_tight, kappa = criterion(output, x.unsqueeze(0), w_real + 1j*w_imag)
		loss_item = loss.item()
		loss_tight.backward()
		optimizer.step()
		losses.append(loss.item())
		kappas.append(kappa)

		if i > max_iter:
			warnings.warn(f"Did not converge after {max_iter} iterations.")
			break
		i += 1

	print(f"Final Fit:\n\tkappa: {kappas[-1]}\n\tloss: {losses[-1]}")
	
	return model.kernels_decoder_real.detach(), model.kernels_decoder_imag.detach(), losses, kappas
